/*
*Copyright(c)2014, Jeffrey Lee
*Allrightsreserved.
*
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
*/
#ifndef CONSTRAINT_H
#define CONSTRAINT_H

#include "common.h"

class constraint
{
private:
	uint32_t refcount;
protected:
	constraint();
	virtual ~constraint();
public:
	virtual cntype gettype() const = 0; /* Get constraint type */

	virtual uint32_t eval(uint32_t opcode) const = 0; /* Evaluate whether constraint matches certain opcode. nonzero = match, zero = no match */

	virtual void emit(const char *opcode) const = 0; /* Emit code to test constraint at runtime */
	virtual void emit_funcs() const = 0; /* Emit any support functions required */

	virtual string describe(const field *fields,uint32_t numfield) const = 0; /* Generate string for displaying constraint in error messages, etc. */

	virtual uint32_t getfields() const = 0; /* Return mask of bits which the constraint result depends on */

	virtual void collect(vector<constraint **> &c) const = 0; /* Collect pointers to all the child constraint pointers for easy deduplication */

	virtual bool operator==(const constraint &c) const = 0; /* Only works properly after constraints have been deduped */

	virtual bool equals(const constraint &c) const = 0; /* Works before constraints have been deduped */

	static constraint *parse(const encoding *e,char *&s);

	static bool deduplicate(constraint **a,constraint **b); /* returns true if b gets deduplicated into a */
	static bool dereference(constraint *c); /* returns true when object gets deleted */
	void reference();

	virtual void treewalk_number(cntype type,int *count,vector<constraint *> &constraints); /* Number things for the tree walker, filling in the given constraint list in order of assignment */

	int treewalk_constraintid;
};

class constraint_field : public constraint
{
private:
	uint8_t startbit;
	uint8_t bitcount;

	~constraint_field();
public:
	constraint_field(const field *f);
	constraint_field(uint32_t start,uint32_t count);

	cntype gettype() const;
	uint32_t eval(uint32_t opcode) const;
	void emit(const char *opcode) const;
	void emit_funcs() const;
	string describe(const field *fields,uint32_t numfield) const;
	uint32_t getfields() const;
	void collect(vector<constraint **> &c) const;
	bool operator==(const constraint &c) const;
	bool equals(const constraint &c) const;

	uint32_t getstartbit() const;
	uint32_t getbitcount() const;
};

class constraint_const : public constraint
{
private:
	uint32_t val;

	~constraint_const();
public:
	constraint_const(uint32_t v);

	cntype gettype() const;
	uint32_t eval(uint32_t opcode) const;
	void emit(const char *opcode) const;
	void emit_funcs() const;
	string describe(const field *fields,uint32_t numfield) const;
	uint32_t getfields() const;
	void collect(vector<constraint **> &c) const;
	bool operator==(const constraint &c) const;
	bool equals(const constraint &c) const;

	uint32_t getvalue() const;
};

class constraint_func : public constraint
{
private:
	funcdef *func;
	constraint **params;
	uint32_t fields; /* Cache sub-constraints constrained field mask for speed */

	~constraint_func();
public:
	constraint_func(funcdef *f,constraint **p);

	cntype gettype() const;
	uint32_t eval(uint32_t opcode) const;
	void emit(const char *opcode) const;
	void emit_funcs() const;
	string describe(const field *fields,uint32_t numfield) const;
	uint32_t getfields() const;
	void collect(vector<constraint **> &c) const;
	bool operator==(const constraint &c) const;
	bool equals(const constraint &c) const;
	void treewalk_number(cntype type,int *count,vector<constraint *> &constraints);

	funcdef *getfunc() const;
	constraint **getparams() const;
};

extern void simplify_and_insert(encoding *e,constraint *c);

#endif

